use crate::{
    data::record::Field,
    error::{ErrorType, MTTError},
};
use std::fmt;
use uuid::Uuid;

#[derive(Clone)]
pub struct ID {
    data: Uuid,
}

impl ID {
    pub fn new(data: Uuid) -> Self {
        Self { data: data }
    }

    pub fn random() -> Self {
        Self {
            data: Uuid::new_v4(),
        }
    }
}

impl TryFrom<Uuid> for ID {
    type Error = MTTError;

    fn try_from(value: Uuid) -> Result<Self, Self::Error> {
        Ok(Self::new(value))
    }
}

impl TryFrom<&str> for ID {
    type Error = MTTError;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match Uuid::try_from(value) {
            Ok(id) => Ok(ID::new(id)),
            Err(_) => Err(MTTError::new(ErrorType::FieldIDInvalid(value.to_string()))),
        }
    }
}

impl fmt::Display for ID {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.data)
    }
}

#[cfg(test)]
mod id {
    use super::*;

    #[test]
    fn try_from_uuid() {
        let data = Uuid::new_v4();
        let id = ID::try_from(data.clone()).unwrap();
        assert_eq!(id.to_string(), data.to_string());
    }

    #[test]
    fn try_from_str() {
        let holder = Uuid::new_v4().to_string();
        let data = holder.as_str();
        let id = ID::try_from(data).unwrap();
        assert_eq!(id.to_string(), data);
    }

    #[test]
    fn try_from_bad_str() {
        let bad = "BadUuid";
        match ID::try_from(bad) {
            Ok(_) => unreachable!("Should have failed to create an ID"),
            Err(err) => match err.get_code() {
                ErrorType::FieldIDInvalid(id) => assert_eq!(id, bad),
                _ => unreachable!("Returned wrong error"),
            },
        }
    }

    #[test]
    fn create_new() {
        let data = Uuid::new_v4();
        let id = ID::new(data.clone());
        assert_eq!(id.to_string(), data.to_string());
    }

    #[test]
    fn random_is_unique() {
        let mut ids: Vec<String> = Vec::new();
        for _ in 0..10 {
            let id = ID::random();
            let holder = id.to_string();
            assert!(!ids.contains(&holder), "'{}' is a duplicate entry", id);
            ids.push(holder);
        }
    }
}
